import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable


# Function for plotting the composition on the materials library
def plot_composition(comps: pd.DataFrame, coords: pd.DataFrame, train_ids: np.ndarray = None, title: str = None,
                     cmap: str = 'viridis'):

    """
    Plots the composition of a materials library.

    :param comps: DataFrame with the compositions to plot
    :param coords: coordinates of each data point
    :param train_ids: indices of areas which should be highlighted
    :param title: title of the plot
    :param cmap: colormap of the scattered points
    :return:
    """

    # If no training IDs were provided, use an empty list
    train_ids = [] if train_ids is None else train_ids - 1
    # Define a list of colors to highlight the training IDs
    e_colors = np.full(342, 'none')
    e_colors[train_ids] = '#C53A32'
    # Create the figure with as many axes as features
    fig, ax = plt.subplots(1, comps.shape[1], figsize=(4 * comps.shape[1], 3.6))
    ax = ax.flatten() if len(ax) > 1 else ax
    # Iterate over each feature
    for i, elem in enumerate(comps.columns):
        # Plot the data points
        ax[i].scatter(coords.iloc[:, 0], coords.iloc[:, 1], c='lightgrey', s=66)
        # Plot the data points
        sc = ax[i].scatter(coords.loc[comps.index].iloc[:, 0], coords.loc[comps.index].iloc[:, 1], c=comps[elem],
                           edgecolors=e_colors, s=66, linewidth=1.5, cmap=cmap)
        # Plot a colorbar
        divider = make_axes_locatable(ax[i])
        cax = divider.append_axes('right', size='5%', pad=0.05)
        fig.colorbar(sc, label='at.%', cax=cax, orientation='vertical')
        # Format the plot
        ax[i].set_title(elem)
        ax[i].set_aspect('equal')
        ax[i].axis('off')
    # Add a title
    fig.suptitle(title, y=1.03)
    # Adjust the space and plot
    fig.tight_layout()
    # Show the plot
    plt.show()


# Function for plotting the correlation of surface and volume composition
def plot_correlation(edx: pd.DataFrame, xps_1: pd.DataFrame, xps_2: pd.DataFrame = None, title: str = None):

    """
    Plots the correlation between the surface and volume composition

    :param edx: DataFrame with the volume compositions
    :param xps_1: DataFrame with the surface compositions
    :param xps_2: (optional second) DataFrame with the volume compositions
    :param title: title of the plot
    :return:
    """

    # Make sure the edx elements are the same as the xps elements
    x_1 = xps_1[edx.columns].div(xps_1[edx.columns].sum(axis=1), axis=0) * 100
    x_2 = xps_2[edx.columns].div(xps_2[edx.columns].sum(axis=1), axis=0) * 100 if xps_2 is not None else None
    # Create the figure with as many axis as elements
    fig, ax = plt.subplots(1, len(edx.columns), figsize=(3.4 * len(edx.columns), 3.4))
    # Plot a title if provided
    if title is not None:
        fig.suptitle(title)
    # Plot every element
    for i, elem in enumerate(edx.columns):
        # First plot the secondary data, so it stays in the background
        if xps_2 is not None:
            ax[i].scatter(edx[elem].loc[x_2.index], x_2[elem], c='lightgrey')
        # Plot the data
        ax[i].scatter(edx[elem].loc[x_1.index], x_1[elem])
        # Plot the correlation line
        xlim, ylim = ax[i].get_xlim(), ax[i].get_ylim()
        ax[i].plot([0, 100], [0, 100], c='grey', linestyle='--')
        ax[i].set_xlim(xlim)
        ax[i].set_ylim(ylim)
        # Set the title as the element
        ax[i].set_title(elem)
        # Add axes labels
        ax[i].set_xlabel('EDX composition (at.%)')
        ax[i].set_ylabel('XPS composition (at.%)')
    # Reduce whitespace
    plt.tight_layout()