import pandas as pd
import numpy as np
import gpflow as gpf


# Predict the XPS composition based on the EDX composition
def predict_xps(edx: pd.DataFrame, xps: pd.DataFrame, coords: pd.DataFrame,
                base_kernel: gpf.kernels.Kernel = None) -> pd.DataFrame:

    """
    Uses multi-output Gaussian process regression to predict the surface composition (XPS) from the volume composition
    (EDX).

    :param edx: DataFrame with the volume composition (input training data)
    :param xps: DataFrame with the surface composition (output training data)
    :param coords: DataFrame with the x and y coordinates for each input training data point
    :param base_kernel: the kernel to use for the Gaussian process, if none is given, the Matérn kernel is used
    :return: DataFrame with the predicted surface composition
    """

    # Define the number of inputs and outputs
    n_inputs = edx.shape[1] + 2  # + 2 for the x-y-coordinates
    n_outputs = xps.shape[1]
    # Prepare the input dataset by combining composition and coordinates
    x_data = pd.concat((coords / 10, edx / 100), axis=1)
    # Standardize it with the common mean and standard deviation
    x_data = (x_data - np.mean(x_data, axis=0)) / np.std(x_data, axis=0)
    # Prepare the output dataset
    y_data = xps / 100

    # Function to prepare the input training data for the Coregionalization kernel
    def augment_x(x: np.ndarray | pd.DataFrame, n_out: int) -> np.ndarray:
        # Augment the input test data with output indices
        output_indices = np.repeat(np.arange(n_out), len(x))
        X_aug = np.hstack((
            np.tile(x, (n_out, 1)),
            output_indices[:, None]
        ))
        # Return the augmented x data
        return X_aug

    # Function to prepare the output training data for the Coregionalization kernel
    def augment_y(y: np.ndarray | pd.DataFrame) -> np.ndarray:
        # Convert the input to a DataFrame if necessary
        y = y.values if isinstance(y, pd.DataFrame) else y
        # Augment the output test data by flattening the output
        return y.T.flatten()[:, None]

    # Define the input test data by augmenting the entire input dataset
    X_test_aug = augment_x(x_data, n_outputs)
    # Define the input training data by augmenting the input data which was also measured by XPS
    X_train_aug = augment_x(x_data.loc[xps.index], n_outputs)
    # For standardizing the output training data, store the mean and standard deviation
    mean_train, std_train = np.mean(y_data, axis=0).values, np.std(y_data, axis=0).values
    # Create the standardized augmented output training data
    y_train_aug = augment_y((y_data - mean_train) / std_train)

    # Define the base kernel
    base_kernel = gpf.kernels.Matern32(active_dims=range(0, n_inputs)) if base_kernel is None else base_kernel
    # Define the coregionalization kernel
    # The active dimension points to the column in the input training data which has the output indices
    coregion_kernel = gpf.kernels.Coregion(output_dim=n_outputs, rank=n_outputs, active_dims=[n_inputs])
    # Combine base and coregionalization kernel
    multioutput_kernel = base_kernel * coregion_kernel

    # Define the model
    model = gpf.models.GPR(data=(X_train_aug, y_train_aug), kernel=multioutput_kernel)
    # Optimize the hyperparameters
    opt = gpf.optimizers.Scipy()
    res = opt.minimize(model.training_loss, model.trainable_variables)
    # Predict the test data
    mean, var = model.predict_y(X_test_aug)
    # Reshape predictions into (n_areas, n_outputs)
    mean = mean.numpy().reshape(n_outputs, len(x_data)).T
    var = var.numpy().reshape(n_outputs, len(x_data)).T
    # De-standardize the outputs
    mean = mean * std_train + mean_train
    var = var * (std_train ** 2)
    # Wrap the predictions into a DataFrame
    mean = pd.DataFrame(mean, columns=y_data.columns, index=range(1, 343))
    var = pd.DataFrame(var, columns=y_data.columns, index=range(1, 343))
    # Normalize the prediction to 100 at.%
    mean = mean.div(mean.sum(axis=1), axis=0) * 100
    # Return the prediction
    return mean
